go :-	clean,
	gen_route,
	gen_incompat,
	gen_endpoints,
	( no_pic_box ; attach_lines_to_boxes),
	tell( 'GbsData.c' ),
	write_header,
	write_decl,
	gen_objects,
	write_incompat,
	write_postsc,
	told.

clean :- retract(route(_,_,_,_)), fail.
clean :- retract(used_by(_,_)), fail.
clean :- retract(incompat(_,_)), fail.
clean :- retract(num_user(_)), fail.
clean :- retract(endpoint(_)), fail.
clean.

append([], Y, Y).
append([A|B], C, [A|D]) :- append(B,C,D).

:- dynamic index/1.
index( 0 ).

gen_endpoints :- pic_point( P, _,_),
		 pic_line( _, P, P0 ),
		 no_other_line( P, P0 ),
		 assert( endpoint( P )), 
		 fail.
gen_endpoints :- pic_point( P, _,_),
		 pic_line( _, P0, P ),
		 no_other_line( P, P0 ),
		 assert( endpoint( P )), 
		 fail.
gen_endpoints.

no_other_line( P, P1 ) :-
		pic_line( _, P, P2 ),
		P1 \= P2, !, fail.
no_other_line( P, P1 ) :-
		pic_line( _, P2, P ),
		P1 \= P2, !, fail.
no_other_line( _, _ ).

attach_lines_to_boxes :-
		pic_box( BX, _,_ ),
		pic_line( LNR, BX, P ),
		attach_line_forw( BX, LNR, P ),
		fail.
attach_lines_to_boxes :-
		pic_box( BX, _,_ ),
		pic_line( LNR, P, BX ),
		attach_line_backw( BX, LNR, P ),
		fail.
attach_lines_to_boxes.

attach_line_forw( BX, LNR, P ) :-
		endpoint( P ),
		assert( used_by( LNR, BX )), !.
attach_line_forw( BX, LNR, P ) :-
		pic_circle( P, _,_ ),
		assert( used_by( LNR, BX )), !.
attach_line_forw( BX, LNR, P ) :-
		assert( used_by( LNR, BX )),
		pic_line( LN1, P, P1),
		attach_line_forw( BX, LN1, P1 ).

attach_line_backw( BX, LNR, P ) :-
		endpoint( P ),
		assert( used_by( LNR, BX )), !.
attach_line_backw( BX, LNR, P ) :-
		pic_circle( P, _,_ ),
		assert( used_by( LNR, BX )), !.
attach_line_backw( BX, LNR, P ) :-
		assert( used_by( LNR, BX )),
		pic_line( LN1, P1, P),
		attach_line_backw( BX, LN1, P1 ).

gen_route :- pic_circle(X, _, _),
	      pic_circle(Y, _, _),
	      X \= Y,
	      pic_line( LID, X, X1 ),
	      ( pic_point( X1, _, _); pic_circle( X1, _,_ )),
	      route_tail( X1, Y, [X, X1], R, [LID], LNR),
	      get_index( IND ),
	      name( IND, EXT ),
	      append( "route", EXT, ROUTE_NAM ),
	      name( ROUTE_AT, ROUTE_NAM ),
	      assert( route( ROUTE_AT, X, Y, R )),
	      mark_used_lines(ROUTE_AT, LNR),
	      fail.
gen_route.

route_tail(X,X,L,R,LN,LN) :- append( L, X, R).
route_tail(X,Y,L,R,LN,LNR) :-
			pic_line( LID, X, Y),
			append(L, [Y], R),
			append(LN, [LID], LNR).
route_tail(X,Y,L,R,LN,LNR) :-
			pic_line( LID, X, P),
		   	pic_point(P, _, _),
			append(L, [P], L1),
			append(LN, [LID], LN1),
		   	route_tail(P, Y, L1, R, LN1, LNR).

get_index( X ) :- retract( index(X) ),
	      	  IND is X + 1,
	      	  assert( index(IND)), !.

mark_used_lines( RA, [ L | TAIL ] ) :-
		assert( used_by( L, RA )),
		mark_used_lines( RA, TAIL ), !.
mark_used_lines( _, [] ).

gen_incompat :- route(R1, S1, Z1, L1 ),
		route(R2, S2, Z2, L2 ),
		check_incompat(R1,S1,Z1,L1, R2,S2,Z2,L2),
		assert( incompat( R1, R2 )),
		fail.
gen_incompat.

check_incompat(  R, _,_,_, R , _,_,_) :- !.
check_incompat( _, S, _,_, _, S, _,_) :- !.
check_incompat( _,_, Z ,_, _,_, Z ,_) :- !.
check_incompat( R1,S1,Z1, [ H | Tail ], R2,S2,Z2, L ) :-
		( inList( H, L ) ;
		check_incompat( R1,S1,Z1, Tail, R2,S2,Z2, L )), !.
check_incompat( _,_,_, [], _,_,_,_ ) :- fail.

inList( E, [ E | _ ] ).
inList( E, [ _ | T ] ) :- inList( E, T ).

:- dynamic coma/1.
coma( 0 ).

no_coma :-	retract( coma(_)), assert( coma( 0 )), !.
need_coma :- 	retract( coma(_)), assert( coma( 1 )), !.
check_coma :-	coma( 1 ), write( ', ' ), !.
check_coma.

no_pic_circle :- assert( pic_circle( dummy, 0, 0 )), fail.
no_pic_circle :- pic_circle( X, _,_ ), X \= dummy,
		 retract( pic_circle( dummy, _,_ )), !, fail.
no_pic_circle :- retract( pic_circle( dummy, _,_ )).

write_circle_decl :- write('x_circle '), pic_circle( NAM, _,_ ),
			check_coma, write( NAM ), need_coma, fail.
write_circle_decl :- no_coma, write(';'), nl, !.

no_pic_box :- assert( pic_box( dummy, 0, 0 )), fail.
no_pic_box :- pic_box( X, _,_ ), X \= dummy,
		 retract( pic_box( dummy, _,_ )), !, fail.
no_pic_box :- retract( pic_box( dummy, _,_ )).

write_box_decl :- write('x_rect '), pic_box( NAM, _,_ ),
			check_coma, write( NAM ), need_coma, fail.
write_box_decl :- no_coma, write(';'), nl, !.

no_pic_point :- assert( pic_point( dummy, 0, 0 )), fail.
no_pic_point :- pic_point( X, _,_ ), X \= dummy,
		 retract( pic_point( dummy, _,_ )), !, fail.
no_pic_point :- retract( pic_point( dummy, _,_ )).

write_point_decl :- write('x_point '), pic_point( NAM, _,_ ),
			check_coma, write( NAM ), need_coma, fail.
write_point_decl :- no_coma, write(';'), nl, !.

no_pic_line :- assert( pic_line( dummy, 0, 0 )), fail.
no_pic_line :- pic_line( X, _,_ ), X \= dummy,
		 retract( pic_line( dummy, _,_ )), !, fail.
no_pic_line :- retract( pic_line( dummy, _,_ )).

write_line_decl :- write('x_line '), pic_line( NAM, _,_ ),
			check_coma, write( NAM ), need_coma, fail.
write_line_decl :- no_coma, write(';'), nl, !.

write_route_decl :- write('route '), route( NAM, _,_,_ ),
			check_coma, write( NAM ), need_coma, fail.
write_route_decl :- no_coma, write(';'), nl, !.

write_decl :-	( no_pic_circle; write_circle_decl ),
		( no_pic_box;	 write_box_decl ),
		( no_pic_point;	 write_point_decl ),
		( no_pic_line;	 write_line_decl),
		write_route_decl.

gen_box_o :- 	pic_box( NAM, XR, YR ), nl, 
		XR1 is XR + 0.5, X is floor( XR1 ),
		YR1 is YR + 0.5, Y is floor( YR1 ),
		write( NAM ), write( ' = Snd2( x_rect_class, new, '),
		write( X ), write( ', ' ), write( Y ), write( ' );'), fail.
gen_box_o :- nl.

gen_circle_o :-	pic_circle( NAM, XR, YR ), nl, 
		XR1 is XR + 0.5, X is floor( XR1 ),
		YR1 is YR + 0.5, Y is floor( YR1 ),
		write( NAM ), write( ' = Snd2( x_circle_class, new, '),
		write( X ), write( ', ' ), write( Y ), write( ' );'), fail.
gen_circle_o :- nl.

gen_point_o :- 	pic_point( NAM, XR, YR ), nl, 
		XR1 is XR + 0.5, X is floor( XR1 ),
		YR1 is YR + 0.5, Y is floor( YR1 ),
		write( NAM ), write( ' = Snd2( x_point_class, new, '),
		write( X ), write( ', ' ), write( Y ), write( ' );'), fail.
gen_point_o :- nl.

gen_line_o :- 	pic_line( NAM, S, Z ), nl, 
		get_num_user( NAM, NUM ),
		write( NAM ), write( ' = Snd' ), 
		NUM1 is NUM + 2,
		write( NUM1 ), write('( x_line_class, new, '),
		write( S ), write( ', ' ), write( Z ), 
		write_user( NAM ),
		write( ' );'), fail.
gen_line_o :- nl.

inc_num_user :- retract( num_user( X )),
		X1 is X + 1,
		assert( num_user( X1 )), !.

get_num_user( _, _ ) :- assert( num_user(0)), fail.
get_num_user( NAM, _ ) :- used_by( NAM, _ ),
			  inc_num_user,
			  fail.
get_num_user( _, NUM ) :- retract( num_user( NUM )), !.

write_user( NAM ) :- used_by( NAM, ROUTE ),
		     write( ', ' ), write( ROUTE ), fail.
write_user( _ ).
		
gen_route_o :- 	route( NAM, X, Y, _ ), nl, 
		write( NAM ), write( ' = Snd2( route_class, new, '),
		write( X ), write( ', ' ), write( Y ), write( ' );'), fail.
gen_route_o :- nl.
		
gen_objects :-	( no_pic_circle; gen_circle_o ),
		( no_pic_box; gen_box_o ),
		( no_pic_point; gen_point_o ),
		gen_route_o,
		( no_pic_line; gen_line_o ).

write_incompat :- route( ROUTE, _, _, _ ),
		  incompat( ROUTE, ROUTE_1 ),
		  nl, write('Snd1( '), write( ROUTE ),
		  write( ', add_incompat, '), write( ROUTE_1 ),
		  write( ' );' ), fail.
write_incompat :- nl.

write_header :-	
	write( '/* GbsData.c - generated by xpic, gbs, gen_route.p */'), nl,
	write( '/* Copyright (c) AP 08-jan-93 */'), nl, nl,
	put("#"), write( 'define __OBJECT_INSTANTIATION'), nl, nl,
	put("#"), write( 'include \"GbsData.h\"'), nl,
	put("#"), write( 'include \"Classes.h\"'), nl,
	put("#"), write( 'include \"XGraClasses.h\"'), nl,
	put("#"), write( 'include \"RouteClasses.h\"'), nl, nl,
	write( 'void init_objs() {'), nl, nl.

write_postsc :- nl, write('}'), nl.
