   /* last(X, L) == X is the last of a list L. */
last(X, [X]).
last(X, [_|Y]) :- last(X, Y).

   /* nextto(X, Y, L) == X is followed by Y in list L. */
nextto(X, Y, [X,Y|_]).
nextto(X, Y, [_|Z]) :- nextto(X, Y, Z).

   /* append(A, B, C) == [A,B] is C */
append([], Y, Y).
append([A|B], C, [A|D]) :- append(B,C,D).

   /* uni-directional: L1 --> L2 */
reverse([], []).
reverse([H|T], L) :- reverse(T, R), append(R, [H], L).

   /* faster implementation */
rev2(L1, L2) :- revzap(L1, [], L2).
revzap([X|L], L2, L3) :- revzap(L, [X|L2], L3).
revzap([], L, L).

   /* efface(X, Y, Z) removes the 1st occruence of element X
      from list Y, giving an "effaced" list Z. */
efface(A, [A|L], L) :- !.
efface(_, [], []).         /* terminate with true */
efface(A, [B|T], [B|L]) :- efface(A, T, L).

   /* delete(X, Y, Z) removes all occruences of element X
      from list Y, giving a "deleted" list Z. */
delete(_, [], []).
delete(X, [X|T], L) :- !, delete(X, T, L).  /* if equal */
delete(X, [H|T], [H|L]) :- delete(X, T, L). /* otherwise */

   /* subst(X, L, X1, L1) substitues all occurences of X in L
      by X1, giving a substituted list L1. */
subst(_, [], _, []).
subst(X, [X|T], X1, [X1|L]) :- !, subst(X, T, X1, L). /* if equal */
subst(X, [H|T], X1, [H|L]) :- subst(X, T, X1, L). /* otherwise */

   /* sublist(L1, L2) asserts that L1 is a sublist of L2. */
sublist([X|L], [X|M]) :- prefix(L, M), ! .
sublist(L, [_|M]) :- sublist(L, M).

   /* prefix(L1, L2) asserts that the beginning of list L2 is list L1. */
prefix([], _).
prefix([X|L], [X|M]) :- prefix(L, M).

   /* position(A, L, P) returns the position of A in a list L if exists; */
   /*                   returns 0 otherwise. */
position(A, L, P) :- posx(A, L, 1, P).

   posx(_, [], _, 0).
   posx(A, [A|_], NextP, NextP) :- !.
   posx(A, [_|T], P, Pos) :- NextP is P+1, posx(A, T, NextP, Pos).

   /* pretty printing of a list */
pp([H|T], I) :- !, J is I+3, pp(H, J), ppx(T, J).
pp(X, I) :- tab(I), write(X), nl.
ppx([], _).
ppx([H|T], I) :- pp(H, I), ppx(T, I).

   /* islist(L) succeeds if L is a list (possibly empty) */
islist([]).
islist([_|_]).
