   /* member(X, Y) == X in Y. */
member(X, [X|_]).
member(X, [_|Y]) :- member(X, Y).

   /* subset(A, B) == A is contained in B. */
subset([], _).
subset([H|T], S) :- member(H, S), subset(T, S).

   /* different(A, B) == A # B. */
different(X, Y) :- not (member(Z, X), member(Z, Y)).

   /* intersection(A, B, C) == A * B = C. */
intersection([], X, []).
intersection([X|R], Y, [X|Z]) :-
   member(X, Y),
   !,
   intersection(R, Y, Z).
intersection([X|R], Y, Z) :-
   intersection(R, Y, Z).

   /* symdifference(A, B, C) with mode(+,+,-) == A | B = C. */
symdifference([], Y, Y).
symdifference([X|R], Y, Z) :-
   exclude(X, Y, Y1),
   !,
   symdifference(R, Y1, Z).
symdifference([X|R], Y, [X|Z]) :-
   symdifference(R, Y, Z).

   exclude(X, [X|R], R).
   exclude(X, [Y|R], [Y|Z]) :- exclude(X, R, Z).

   /* union(A, B, C) == A + B = C. */
union([], X, X).
union([X|R], Y, Z) :-
   member(X, Y),
   !,
   union(R, Y, Z).
union([X|R], Y, [X|Z]) :-
   union(R, Y, Z).

   /* difference(A, B, C) == A - B = C. */
difference([], _, []).
difference([H|T], A, [H|R]) :-
   not member(H, A),
   !,
   difference(T, A, R).
difference([H|T], A, R) :-
   difference(T, A, R).
